/*
 * vip1710player.c: 
 *
 * See the README file for copyright information and how to reach the author.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <sys/select.h>

#include "vip1710player.h"
#include "vip1710tools.h"

#define DLOG(...) log(vipDEBUG1, "vip1710/player: " __VA_ARGS__)
#define ELOG(...) log(vipERROR, "vip1710/player: " __VA_ARGS__)

extern "C" {
/* interface definitions: */
IPCPROTO_NEW_SERVER(Vip1710PlayerService, "VDRPlayerService")
IPCPROTO_NEW_PARSER(Vip1710PlayerService, SetVolume, "i")
IPCPROTO_NEW_PARSER(Vip1710PlayerService, SetWSS, "b")
IPCPROTO_NEW_PARSER(Vip1710PlayerService, SetVideoWindow, "iiii")
IPCPROTO_NEW_PARSER(Vip1710PlayerService, SetVideoFrame, "iiii")

IPCPROTO_NEW_GENERIC(Vip1710PlayerServiceCaller, "VDRPlayerService")
IPCPROTO_NEW_CALL(Vip1710PlayerServiceCaller, Quit, "P!")
IPCPROTO_NEW_CALL(Vip1710PlayerServiceCaller, Play, "P!")
IPCPROTO_NEW_CALL(Vip1710PlayerServiceCaller, Pause, "P!")
IPCPROTO_NEW_CALL(Vip1710PlayerServiceCaller, Stop, "P!")
IPCPROTO_NEW_CALL(Vip1710PlayerServiceCaller, SetVolume, "iP!")
IPCPROTO_NEW_CALL(Vip1710PlayerServiceCaller, SetWSS, "bP!")
IPCPROTO_NEW_CALL(Vip1710PlayerServiceCaller, SetVideoWindow, "iiiiP!")
IPCPROTO_NEW_CALL(Vip1710PlayerServiceCaller, SetVideoFrame, "iiiiP!")

IPCPROTO_NEW_GENERIC(DummyAppServiceCaller, "IfaceDummyAppService")
IPCPROTO_NEW_CALL(DummyAppServiceCaller, GetAppId, "P|Ei")
}

typedef enum {
  NoPlayer,
  Creating,
  Closed,
  Open,
  Playing,
  Paused,
  Quitting
} ePlayState;

cVip1710Player::cVip1710Player(void): cThread("VIP1710 player thread")
{
	playerservicecaller = NULL;
	Start();
}

cVip1710Player::~cVip1710Player() 
{
	Stop();
}

void cVip1710Player::Stop(void) 
{
	if (Running()) {
		Quit();
		Cancel(3);
	}
}

void cVip1710Player::Action(void)
{
	int err;
	int application_id;

	int vipstate;
	int position1;
	int position2;
	int volume;
	ePlayState current_state;
	ePlayState target_state;
	bool volume_changed = 0;
	int wss;
	bool wss_changed = 0;
	int vw_x, vw_y, vw_width, vw_height;
	bool vw_changed = 0;
	int vf_x, vf_y, vf_width, vf_height;
	bool vf_changed = 0;

	int c;

	ipcconn playerservice = NULL;
	ipcconn dummyapp = NULL;
	ipcconn nameservice = NULL;
	ipcconn mediaservice = NULL;
	ipcconn mediaplayer = NULL;
	ipcconn videomixerservice = NULL;
	ipcconn videomixer = NULL;
	ipcconn volumeservice = NULL;
	ipcconn playerserviceclient[VIP1710_IPC_CLIENTS];

	for(c = 0; c < VIP1710_IPC_CLIENTS; c++)
		playerserviceclient[c] = NULL;

	playerservice = Vip1710PlayerService_new(VIP1710_PLUGIN_SOCKET);
	if(!playerservice) {
		ELOG("cannot create PlayerService");
		goto quit;
	}
	dummyapp = DummyAppServiceCaller_new(VIP1710_DUMMY_APP_SOCKET);
	if(!dummyapp) {
		ELOG("cannot create DummyAppServiceCaller");
		goto quit;
	}
	nameservice = ipcconn_get_nameservice();
	if(!nameservice) {
		ELOG("cannot create NameService caller");
		goto quit;
	}
	mediaservice = MediaService_new(nameservice);
	if(!mediaservice) {
		ELOG("cannot create MediaService caller");
		goto quit;
	}
	volumeservice = VolumeService_new(nameservice);
	if(!volumeservice) {
		ELOG("cannot create VolumeService caller");
		goto quit;
	}
	err = ipcconn_bind_and_listen(playerservice, 0777, VIP1710_IPC_CLIENTS);
	if(err) {
		ELOG("unable bind and listen on socket %s", VIP1710_PLUGIN_SOCKET);
		goto quit;
	}
	err = DummyAppServiceCaller_GetAppId(dummyapp, &application_id);
	if(err) {
		ELOG("unable to get application ID from dummyapplication. Is it really running?");
		goto quit;
	}
	err = VolumeService_GetVolume(volumeservice, &volume);
	if(err) {
		ELOG("cannot get current volume setting from VolumeService");
		goto quit;
	}
	videomixerservice = VideoMixerService_new(nameservice);
	if(!videomixerservice) {
		ELOG("cannot create VideoMixerService");
		goto quit;
	} else {
		char *videomixersocket = NULL;
		err = VideoMixerService_GetInstance(videomixerservice, application_id, &videomixersocket);
		videomixer = VideoMixer_new(videomixersocket);
		if(!videomixer) {
			ELOG("cannot create VideoMixer");
			goto quit;
		}
		VideoMixer_SetFullVideoWindow(videomixer);
		VideoMixer_SetFullVideoFrame(videomixer);
		VideoMixer_EnableWideScreen(videomixer, 0);
		VideoMixer_EnableVideoWindow(videomixer, 1);
	}

	current_state = NoPlayer;
	target_state = Closed;

	while(Running()) {
		int ret;
		int nfds;
		fd_set rfds;
		struct timeval tv;

		/* state machine: */
		if(target_state != current_state) {
			DLOG("state change requested: %d -> %d", current_state, target_state);
			if(current_state == NoPlayer) {
				if(target_state != Quitting) {
					err = MediaService_CreatePlayerInstance(mediaservice, application_id, VIP1710_PLUGIN_SOCKET);
					if(!err) {
						current_state = Creating;
					}
					DLOG("creating player, result: %d (%s)", err, err ? "error" : "OK");
				}
			}
			if(current_state == Closed) {
				if(target_state == Playing) {
					err = MediaPlayer_Open(mediaplayer, "lstrm://" VIP1710_TS_SOCKET);
					if(!err)
						current_state = Open;
					DLOG("stream Open, result: %d (%s)", err, err ? "error" : "OK");
				}
			}
			if(current_state == Open) {
				if(target_state == Playing || target_state == Paused) {
					err = MediaPlayer_Play(mediaplayer);
					if(!err)
						current_state = Playing;
					DLOG("play command, result: %d (%s)", err, err ? "error" : "OK");
				}
			}
			if(current_state == Open || current_state == Playing || current_state == Paused) {
				if(target_state == Closed || target_state == Quitting) {
					err = MediaPlayer_Close(mediaplayer);
					if(!err || target_state == Quitting)
						current_state = Closed;
					DLOG("close command, result: %d (%s)", err, err ? "error" : "OK");
				}
			}
			if(current_state == Playing) {
				if(target_state == Paused) {
					err = MediaPlayer_Pause(mediaplayer);
					if(!err)
						current_state = Paused;
					DLOG("paused command, result: %d (%s)", err, err ? "error" : "OK");
				}
			}
			if(current_state == Paused) {
				if(target_state == Playing) {
					err = MediaPlayer_Play(mediaplayer);
					if(!err)
						current_state = Playing;
					DLOG("play command, result: %d (%s)", err, err ? "error" : "OK");
				}
			}
			if(current_state == Closed) {
				if(target_state == Quitting) {
					err = MediaPlayer_ReleaseInstance(mediaplayer);
					current_state = NoPlayer;
					ipcconn_free(&mediaplayer);
					DLOG("releasing player instance, result: %d (%s)", err, err ? "error, ignoring!" : "OK");
				}
			}
			if(current_state == NoPlayer || current_state == Creating) {
				if(target_state == Quitting) {
					current_state = Quitting;
					break; /* break out of Running() loop */
				}
			}
		}
		if(volume_changed) {
			err = VolumeService_SetVolume(volumeservice, volume);
			volume_changed = 0;
		}
		if(wss_changed) {
			err = VideoMixer_EnableWideScreen(videomixer, wss);
			wss_changed = 0;
		}
		if(vw_changed) {
			err = VideoMixer_SetVideoWindow(videomixer, vw_x, vw_y, vw_width, vw_height);
			vw_changed = 0;
		}
		if(vf_changed) {
			err = VideoMixer_SetVideoWindow(videomixer, vf_x, vf_y, vf_width, vf_height);
			vf_changed = 0;
		}

		/* select() timeout: 1 second */
		tv.tv_sec = 1;
		tv.tv_usec = 0;

		FD_ZERO(&rfds);
		FD_SET(playerservice->fd, &rfds);
		nfds = playerservice->fd;
		for(c = 0; c < VIP1710_IPC_CLIENTS; c++)
			if(playerserviceclient[c]) {
				FD_SET(playerserviceclient[c]->fd, &rfds);
				if(playerserviceclient[c]->fd > nfds)
					nfds = playerserviceclient[c]->fd;
			}
		nfds++;

		ret = select(nfds, &rfds, NULL, NULL, &tv);
		if(ret == -1) {
			ELOG("fatal error during select(), exiting");
			break;
		} else if (ret) {
			/* events on our socket(s) */
			for(c = 0; c < VIP1710_IPC_CLIENTS; c++) if(playerserviceclient[c] && FD_ISSET(playerserviceclient[c]->fd, &rfds)) {
				/* incoming data on client connection */
				char *procedure = NULL;
				char *interface = NULL;
				int status;
				ipcmessage msg;
				err = ipcconn_receive_rpc(playerserviceclient[c], &msg, &status, &interface, &procedure);
				if(!err) {
					/* dispatch... */
					DLOG("got call for interface %s...", interface);
					if(!MediaPlayerObserver_check_interface(interface)) {
						if(!strcmp("OnMediaPlayerInitialized", procedure)) {
							char *mediaplayersocket = NULL;
							err = MediaPlayerObserver_parse_OnMediaPlayerInitialized(msg, &mediaplayersocket);
							if(!err) {
								mediaplayer = MediaPlayer_new(mediaplayersocket);
								if(!mediaplayer) {
									ELOG("cannot create MediaPlayer, deadlock situation, retrying");
									current_state = NoPlayer;
									sleep(1); /* probably not a good idea */
								} else {
									err = MediaPlayer_Activate(mediaplayer);
									current_state = Closed;
								}
								free(mediaplayersocket);
							}
							DLOG("procedure %s called, action result is %d", procedure, err);
						} else {
							DLOG("MPO: unknown procedure %s called", procedure);
						}
					} else if(!MediaPlayerObserverBase_check_interface(interface)) {
						if(!strcmp("OnStateChanged", procedure)) {
							char *statename = NULL;
							err = MediaPlayerObserverBase_parse_OnStateChanged(msg, &statename, &vipstate);
							if(!err) {
								DLOG("state change reported, new state %d (%s)", vipstate, statename);
								free(statename);
							}
							DLOG("procedure %s called, action result is %d", procedure, err);
						} else if(!strcmp("OnPositionChanged", procedure)) {
							err = MediaPlayerObserverBase_parse_OnPositionChanged(msg, &position1, &position2);
							DLOG("procedure %s called, action result is %d", procedure, err);
						} else if(!strcmp("OnDataAvailable", procedure)) {
							int value;
							err = MediaPlayerObserverBase_parse_OnDataAvailable(msg, &value);
							DLOG("procedure %s called, action result is %d", procedure, err);
						} else {
							DLOG("MPOB: unknown procedure %s called", procedure);
						}
					} else if(!Vip1710PlayerService_check_interface(interface)) {
						if(!strcmp("Quit", procedure)) {
							target_state = Quitting;
						} else if(!strcmp("Play", procedure)) {
							target_state = Playing;
						} else if(!strcmp("Pause", procedure)) {
							target_state = Paused;
						} else if(!strcmp("Stop", procedure)) {
							target_state = Closed;
						} else if(!strcmp("SetVolume", procedure)) {
							if((volume >= 0) && !Vip1710PlayerService_parse_SetVolume(msg, &volume))
								volume_changed = true;
						} else if(!strcmp("SetWSS", procedure)) {
							if(!Vip1710PlayerService_parse_SetWSS(msg, &wss))
								wss_changed = true;
						} else if(!strcmp("SetVideoWindow", procedure)) {
							if(!Vip1710PlayerService_parse_SetVideoWindow(msg, &vw_height, &vw_width, &vw_y, &vw_x))
								vw_changed = true;
						} else if(!strcmp("SetVideoFrame", procedure)) {
							if(!Vip1710PlayerService_parse_SetVideoFrame(msg, &vf_height, &vf_width, &vf_y, &vf_x))
								vf_changed = true;
						}
					}
					ipcmessage_free(&msg);
					free(interface);
					free(procedure);
				} else {
					ipcconn_free(&playerserviceclient[c]);
				}
			} /* FD_ISSET(playerserviceclient[c], &rfds) */

			if(FD_ISSET(playerservice->fd, &rfds)) {
				/* connection pending on server socket */
				for(c = 0; c < VIP1710_IPC_CLIENTS; c++)
					if(!playerserviceclient[c]) {
						ipcconn_accept(playerservice, &playerserviceclient[c]);
						break;
					}
				if(c == VIP1710_IPC_CLIENTS) {
					ipcconn dropper;
					ELOG("error: no free client slot for incoming connection.");
					ipcconn_accept(playerservice, &dropper);
					ipcconn_free(&dropper);
				}
			}
		} /* ret from select() */
	} /* while(Running()) */
	
	for(c = 0; c < VIP1710_IPC_CLIENTS; c++)
		if(playerserviceclient[c])
			ipcconn_free(&playerserviceclient[c]);

quit:
	ipcconn_free(&videomixer);
	ipcconn_free(&videomixerservice);
	ipcconn_free(&volumeservice);
	ipcconn_free(&mediaservice);
	ipcconn_free(&nameservice);
	ipcconn_free(&playerservice);
}

bool cVip1710Player::Connect(void)
{
	if(playerservicecaller)
		return true;
	playerservicecaller = Vip1710PlayerServiceCaller_new(VIP1710_PLUGIN_SOCKET);
	if(playerservicecaller) {
		int err = ipcconn_connect(playerservicecaller);
		if(!err)
			return true;
	}
	return false;
}

void cVip1710Player::Quit(void)
{
	if(Connect()) {
		Vip1710PlayerServiceCaller_Quit(playerservicecaller);
	}
}

void cVip1710Player::VidPlay(void)
{
	if(Connect()) {
		Vip1710PlayerServiceCaller_Play(playerservicecaller);
	}
}

void cVip1710Player::VidPause(void)
{
	if(Connect()) {
		Vip1710PlayerServiceCaller_Pause(playerservicecaller);
	}
}

void cVip1710Player::VidStop(void)
{
	if(Connect()) {
		Vip1710PlayerServiceCaller_Stop(playerservicecaller);
	}
}

void cVip1710Player::SetVolume(int volume)
{
	if(Connect()) {
		Vip1710PlayerServiceCaller_SetVolume(playerservicecaller, volume);
	}
}

void cVip1710Player::SetAspectRatio(int aspectratio, bool wss)
{
	if(!wss) {
		// this assumes square pixels (which is probably wrong) and PAL
		if(Connect()) {
			Vip1710PlayerServiceCaller_SetWSS(playerservicecaller, 0);
			switch (aspectratio) {
			case 1:  /* 1:1 aspect ratio */
				Vip1710PlayerServiceCaller_SetVideoWindow(playerservicecaller, 72, 0, 576, 576);
				//Vip1710PlayerServiceCaller_SetVideoFrame(playerservicecaller, 0, 0, 720, 576);
				break;
			case 2:  /* 4:3  aspect ratio */
				Vip1710PlayerServiceCaller_SetVideoWindow(playerservicecaller, 0, 0, 720, 576);
				//Vip1710PlayerServiceCaller_SetVideoFrame(playerservicecaller, 0, 0, 720, 576);
				break;
			case 3:  /* 16:9 aspect ratio, currently letterbox */
				Vip1710PlayerServiceCaller_SetVideoWindow(playerservicecaller, 0, 85, 720, 405);
				//Vip1710PlayerServiceCaller_SetVideoFrame(playerservicecaller, 0, 0, 720, 576);
				break;
			case 4: /* 2:21:1 aspect ratio, currently letterbox */
				Vip1710PlayerServiceCaller_SetVideoWindow(playerservicecaller, 0, 125, 720, 326);
				//Vip1710PlayerServiceCaller_SetVideoFrame(playerservicecaller, 0, 0, 720, 576);
				break;
			case 5 ... 15:
				break;
			default:
				break;
			}
		}
	} else {
		// this assumes 16:9 / 4:3 PAR = 1:1.333
		switch (aspectratio) {
		case 1:  /* 1:1 aspect ratio */
			Vip1710PlayerServiceCaller_SetWSS(playerservicecaller, 0);
			Vip1710PlayerServiceCaller_SetVideoWindow(playerservicecaller, 72, 0, 576, 576);
			//Vip1710PlayerServiceCaller_SetVideoFrame(playerservicecaller, 0, 0, 720, 576);
			break;
		case 2:  /* 4:3  aspect ratio */
			Vip1710PlayerServiceCaller_SetWSS(playerservicecaller, 0);
			Vip1710PlayerServiceCaller_SetVideoWindow(playerservicecaller, 0, 0, 720, 576);
			//Vip1710PlayerServiceCaller_SetVideoFrame(playerservicecaller, 0, 0, 720, 576);
			break;
		case 3:  /* 16:9 aspect ratio */
			Vip1710PlayerServiceCaller_SetWSS(playerservicecaller, 1);
			Vip1710PlayerServiceCaller_SetVideoWindow(playerservicecaller, 0, 0, 720, 576);
			//Vip1710PlayerServiceCaller_SetVideoFrame(playerservicecaller, 0, 0, 720, 576);
			break;
		case 4: /* 2:21:1 aspect ratio, currently letterbox */
			Vip1710PlayerServiceCaller_SetWSS(playerservicecaller, 1);
			Vip1710PlayerServiceCaller_SetVideoWindow(playerservicecaller, 0, 94, 720, 480);
			//Vip1710PlayerServiceCaller_SetVideoFrame(playerservicecaller, 0, 0, 720, 576);
			break;
		case 5 ... 15:
			break;
		default:
			break;
		}
	}
}

